@@ -3,14 +3,15 @@
 #else
 #	include <dlfcn.h>
 #endif
-#include "libHX.h"
+#include <libHX/misc.h>
+#include "internal.h"
 
 EXPORT_SYMBOL void *HX_dlopen(const char *file)
 {
 #ifdef _WIN32
 	return LoadLibrary(file);
 #else
-	return dlopen(file, RTLD_NOW); 
+	return dlopen(file, RTLD_LAZY); 
 #endif
 }
 
