@@ -1,6 +1,6 @@
 /*
  *	libHX/string.c
- *	Copyright © Jan Engelhardt <jengelh [at] gmx de>, 1999 - 2008
+ *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 1999 - 2008
  *
  *	This file is part of libHX. libHX is free software; you can
  *	redistribute it and/or modify it under the terms of the GNU
@@ -8,11 +8,11 @@
  *	Foundation; either version 2.1 or 3 of the License.
  */
 #include <sys/types.h>
-#include <ctype.h>
 #include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <libHX/ctype_helper.h>
 #include <libHX/string.h>
 #include "internal.h"
 
@@ -25,8 +25,36 @@ EXPORT_SYMBOL char *HX_basename(const char *s)
 {
 	const char *p;
 	if ((p = strrchr(s, '/')) != NULL)
-		return const_cast(char *, p + 1);
-	return const_cast(char *, s);
+		return const_cast1(char *, p + 1);
+	return const_cast1(char *, s);
+}
+
+EXPORT_SYMBOL char *HX_basename_exact(const char *s)
+{
+	const char *start, *end;
+	char *ret;
+	int len;
+
+	if (*s == '\0')
+		return HX_strdup(".");
+	for (end = s + strlen(s) - 1; end >= s && *end == '/'; --end)
+		;
+	if (end < s)
+		return HX_strdup("/");
+
+	start = HX_strbchr(s, end, '/');
+	if (start == NULL) {
+		len = end - s + 1;
+		ret = HX_memdup(s, len + 1);
+	} else {
+		++start;
+		len = end - start + 1;
+		ret = HX_memdup(start, len + 1);
+	}
+	if (ret == NULL)
+		return NULL;
+	ret[len] = '\0';
+	return ret;
 }
 
 EXPORT_SYMBOL char *HX_chomp(char *s)
@@ -131,6 +159,28 @@ EXPORT_SYMBOL char **HX_split(const char *str, const char *delim,
 	return ret;
 }
 
+EXPORT_SYMBOL char **HX_split4(char *s, const char *delim, int *fld, int max)
+{
+	char **stk;
+	const char *p = s;
+	int count = 1;
+
+	for (p = strpbrk(p, delim); p != NULL; p = strpbrk(++p, delim))
+		if (++count >= max && max > 0) {
+			count = max;
+			break;
+		}
+
+	stk = malloc(sizeof(char *) * (count + 1));
+	if (stk == NULL)
+		return NULL;
+	stk[count] = NULL;
+	count = HX_split5(s, delim, count, stk);
+	if (fld != NULL)
+		*fld = count;
+	return stk;
+}
+
 EXPORT_SYMBOL int HX_split5(char *s, const char *delim, int max, char **stk)
 {
 	/*
@@ -186,7 +236,7 @@ EXPORT_SYMBOL char *HX_strlower(char *expr)
 {
 	char *orig = expr;
 	while (*expr != '\0') {
-		*expr = tolower(*expr);
+		*expr = HX_tolower(*expr);
 		++expr;
 	}
 	return orig;
@@ -198,7 +248,7 @@ EXPORT_SYMBOL size_t HX_strltrim(char *expr)
 	size_t diff = 0;
 	travp = expr;
 
-	while (*travp != '\0' && isspace(*travp))
+	while (*travp != '\0' && HX_isspace(*travp))
 		++travp;
 	if ((diff = travp - expr) > 0)
 		memmove(expr, travp, diff);
@@ -212,9 +262,7 @@ EXPORT_SYMBOL char *HX_strmid(const char *expr, long offset, long length)
 
 	if (offset < 0)
 		offset = strlen(expr) + offset;
-	if (length == 0)
-		length = strlen(expr) - offset;
-	else if (length < 0)
+	if (length < 0)
 		length = strlen(expr) - offset + length;
 	if ((buffer = malloc(length + 1)) == NULL)
 		return NULL;
@@ -233,23 +281,24 @@ EXPORT_SYMBOL size_t HX_strrcspn(const char *s, const char *rej)
 	return n;
 }
 
-EXPORT_SYMBOL char *HX_strrev(char *expr)
+EXPORT_SYMBOL char *HX_strrev(char *s)
 {
-	char *dyn = HX_strdup(expr), *orig = expr;
-	size_t s = strlen(dyn);
+	size_t i, z = strlen(s)-1, z2 = z / 2;
 
-	dyn += s;
-	while (s--)
-		*expr++ = *--dyn;
+	for (i = 0; i < z2; ++i) {
+		char temp;
+		temp = s[i];
+		s[i] = s[z-i];
+		s[z-i] = temp;
+	}
 
-	free(dyn);
-	return orig;
+	return s;
 }
 
 EXPORT_SYMBOL size_t HX_strrtrim(char *expr)
 {
 	int i = strlen(expr), s = 0;
-	while (i-- && isspace(expr[i]))
+	while (i-- && HX_isspace(expr[i]))
 		++s;
 	expr[++i] = '\0';
 	return s;
@@ -303,7 +352,7 @@ EXPORT_SYMBOL char *HX_strupper(char *expr)
 {
 	char *orig = expr;
 	while (*expr != '\0') {
-		*expr = toupper(*expr);
+		*expr = HX_toupper(*expr);
 		++expr;
 	}
 	return orig;
